#!/usr/bin/perl 
#
# This script encodes the text into UTF.
#
# Hindi letters in the text need to be typed in the kannada alphabet as 
# per our convention and ***the "letters" should be seperated by space***.
# Matras are written as c[2-9]. ex: rani is written as r c2 n c3.		
# Matra 1 for with the short vowel 'a' should not be used. A consonant 
# with out a halant following it, is  impliciltely interpreted as one 
# with matra 1 following it.
# For entering the halves, all the consonants preceding the matra(if there it's not 'a') in the half,
# should contain a halant excepting the last consonant which would be a CV.
#  ex: rashtra is written as,  r c2 sh HH tt HH r.  (HH for halant as mentioned below)
#
# Map for other symbols (ascii characters) that we use in our conventional alphabet	
#  is as follows.
#
# 	space is SS
# 	Halant is HH
# 	period is PP
#	comma  is CC
#
# The output UTF will be written in ./utf


%map=("an",0x0901,":",0x0902,"1",0x0905,"2",0x0906,"3",0x0907,"4",0x0908,"5",0x0909,"6",0x090A,"7",0x090E,"8",0x090F,"9",0x0910,"10",0x0911,"11",0x0912,"12",0x0913,"13",0x0914,"k",0x0915,"kh",0x0916,"g",0x0917,"gh",0x0918,"nga",0x0919,"ch",0x091A,"chh",0x091B,"j",0x091C,"jh",0x091D,"nja",0x091E,"tt",0x091F,"tth",0x0920,"dd",0x0921,"ddh",0x0922,"nna",0x0923,"t",0x0924,"th",0x0925,"d",0x0926,"dh",0x0927,"n",0x0928,"p",0x092A,"f",0x092B,"b",0x092C,"bh",0x092D,"m",0x092E,"y",0x092F,"r",0x0930,"l",0x0932,"ll",0x0934,"v",0x0935,"sh",0x0936,"s",0x0938,"h",0x0939,"c2",0x093E,"c3",0x093F,"c4",0x0940,"c5",0x0941,"c6",0x0942,"c7",0x0946,"c8",0x0947,"c9",0x0948,"c11",0x094A,"c12",0x094B,"c13",0x094C,"HH",0x094D,"CC"," ","PP",0x0964,"SS",0x0965,"n0",0x0966,"n1",0x0967,"n2",0x0968,"n3",0x0969,"n4",0x096A,"n5",0x096B,"n6",0x096C,"n7",0x096D,"n8",0x096E,"n9",0x096F);

open(OUT,"> utf");

	
sub uencode{
        $code  = 0xE08080;
        ($c)=@_;
        if( ($c eq " ") || ($c eq ".") || ($c eq ",") ){
                $code =  $_;
                }
        else{
                $databyte1 = $c >> 12;
                $byte1 =  $databyte1 << 16;
                $code  =  $code | $byte1;

                $databyte2 = $c >> 6;          
                $databyte2 = $databyte2 & 0x003F;
                $databyte2 = $databyte2 << 8;
                $code = $code | $databyte2;

                $databyte3 = $c & 0x003F ;
                $code = $code | $databyte3;
                }
                return($code);
        }


while(1){
        $_ = <STDIN>;
        chomp;
        if(/stop/){last}
        @symbs = split " " ;
        $vf=0;
        foreach(@symbs){
                if(/#/){
                        syswrite(OUT, $_);
                        if(!$vf){ $vf=1; }
                        else{ $vf=0; }
                        }
                elsif ($vf){ syswrite( OUT , $_." " ); }
                else{
                        $cc = uencode($map{$_});
                        my ($c1,$c2,$c3);
                        $c1 =  $cc & (0x000000ff);
                        $c2 = ($cc & (0x0000ff00)) >> 8;
                        $c3 = ($cc & (0x00ff0000)) >> 16;
                        $a=pack("C3",$c3,$c2,$c1);
                        syswrite(OUT,$a);
                     }
        }
}                            

